@@ -45,6 +45,7 @@ static char *configpath = NULL;
 
 /** The command line arguments. */
 static char *target = NULL;
+static char *remote_user = NULL;
 static char *cwd = NULL;
 static char **args = NULL;
 static int action = PTYPE_COMMAND;
@@ -461,6 +462,7 @@ static void usage(void)
 	fprintf(stderr, "Usage: %s [-t|--target <target>]"
 				 " [-c|--config <path>]"
 				 " [-d|--directory <dir>]"
+				 " [-r|--remote <uid>,<gid>,<username>]"
 				 " [<command>]"
 				 " [<args>]\n"
 			"       %s [-t|--target <target>]"
@@ -557,6 +559,7 @@ static void read_args(int orig_argc, char **orig_argv)
 		{ "target",     required_argument, 0, 't' },
 		{ "config",     required_argument, 0, 'c' },
 		{ "directory",  required_argument, 0, 'd' },
+		{ "remote",     required_argument, 0, 'r' },
 		{ "mount",      no_argument,       &action, PTYPE_MOUNT  },
 		{ "umount",     no_argument,       &action, PTYPE_UMOUNT },
 		{ 0 }
@@ -567,7 +570,7 @@ static void read_args(int orig_argc, char **orig_argv)
 	while (1) {
 		int c;
 
-		c = getopt_long(argc, argv, "hvt:c:d:", longopts, NULL);
+		c = getopt_long(argc, argv, "hvt:c:d:r:", longopts, NULL);
 		if (c < 0) {
 			break;
 		}
@@ -585,6 +588,10 @@ static void read_args(int orig_argc, char **orig_argv)
 			cwd = optarg;
 			break;
 
+		case 'r':
+			remote_user = optarg;
+			break;
+
 		case 0:
 			/* --mount or --umount */
 			break;
@@ -630,6 +637,62 @@ static void cleanup(void)
 	debug("sbrsh exiting");
 }
 
+static int determine_user_info(uid_t *uidp, gid_t *gidp, char *username)
+{
+	uid_t uid;
+	gid_t gid;
+
+	if (remote_user != NULL) {
+		/*
+		 * Parse the list for uid, gid, and username (string placed at
+		 * end for simplicity)
+		 */
+		if (sscanf(remote_user, "%d,%d,%s", &uid, &gid, username) != 3) {
+			error("Invalid format for -r argument: %s", remote_user);
+			return 1;
+		}
+
+	} else {
+		char *str;
+
+		uid = geteuid();
+		gid = getegid();
+
+		struct passwd *userstruct = getpwuid(uid);
+		if (!userstruct) {
+			error("Can't get user information about uid %d", uid);
+			return 1;
+		}
+
+		strcpy(username, userstruct->pw_name);
+
+		/*
+		 * Check for environment overrides
+		 */
+		str = getenv("_SBOX_NONFAKE_USER");
+		if (str) {
+			strcpy(username, str);
+		}
+
+		str = getenv("_SBOX_NONFAKE_UID");
+		if (str && (uid = atoi(str)) <= 0) {
+			error("Invalid _SBOX_NONFAKE_UID: %d", uid);
+			return 1;
+		}
+
+		str = getenv("_SBOX_NONFAKE_GID");
+		if (str && (gid = atoi(str)) <= 0) {
+			error("Invalid _SBOX_NONFAKE_GID: %d", gid);
+			return 1;
+		}
+	}
+
+	*uidp = uid;
+	*gidp = gid;
+
+	return 0;
+}
+
 /*
  * Reads options, resolves and connects to host, talks with it, returns rc.
  */
@@ -638,8 +701,9 @@ int main(int argc, char **argv)
 	config_t *cfg;
 	mount_info_t **mounts;
 	int16_t rc;
-	uid_t real_uid;
-	gid_t real_gid;
+	uid_t uid = 0;
+	gid_t gid = 0;
+	char user[NAME_MAX];
 
 	progname = get_progname(argv[0]);
 
@@ -735,7 +799,8 @@ int main(int argc, char **argv)
 			debug("type=%d device=%s point=%s opts=%s",
 			      mi->type, mi->device, mi->point, mi->opts);
 
-			if (action == PTYPE_COMMAND && MTYPE_NFS == mi->type) {
+			if (action == PTYPE_COMMAND && (MTYPE_NFS == mi->type ||
+			                                MTYPE_SSH == mi->type)) {
 				if (mntinfo_stat_device(mi) < 0)
 					return 1;
 
@@ -810,39 +875,19 @@ int main(int argc, char **argv)
 	}
 
 	/*
-	 * Find out user and group ID
+	 * Find out remote user ID, group ID, and username
 	 */
-	real_uid = geteuid();
-	real_gid = getegid();
-
-	if (real_uid == 0 || real_gid == 0) {
-		char *str;
-
-		str = getenv("_SBOX_NONFAKE_UID");
-		if (str && (real_uid = atoi(str)) <= 0) {
-			error("Invalid _SBOX_NONFAKE_UID: %d", real_uid);
-			return 1;
-		}
-
-		str = getenv("_SBOX_NONFAKE_GID");
-		if (str && (real_gid = atoi(str)) <= 0) {
-			error("Invalid _SBOX_NONFAKE_GID: %d", real_gid);
-			return 1;
-		}
+	if (determine_user_info(&uid, &gid, user)) {
+		return 1;
 	}
 
 	/*
 	 * Send user name
 	 */
 	{
-		struct passwd *userstruct = getpwuid(real_uid);
-		if (!userstruct) {
-			error("Can't get user information about uid %d", real_uid);
-			return 1;
-		}
 
-		debug("Sending USER packet: %s", userstruct->pw_name);
-		if (write_str_packet(sd, PTYPE_USER, userstruct->pw_name) < 0) {
+		debug("Sending USER packet: %s", user);
+		if (write_str_packet(sd, PTYPE_USER, user) < 0) {
 			error("Can't send USER packet");
 			return 1;
 		}
@@ -963,17 +1008,23 @@ int main(int argc, char **argv)
 		}
 
 		/*
-		 * Send effective user and group ID and supplementary group IDs
+		 * Send effective user and group ID.  If the user was not
+		 * faked, then send supplementary group IDs; otherwise
+		 * discover these on the remote side.
 		 */
 		{
 			gid_t gids[NGROUPS_MAX];
 			int num, i;
 			uint16v_t *ids;
 
-			num = getgroups(NGROUPS_MAX, gids);
-			if (num < 0) {
-				error("Can't get supplementary group IDs");
-				return 1;
+			if (!remote_user) {
+				num = getgroups(NGROUPS_MAX, gids);
+				if (num < 0) {
+					error("Can't get supplementary group IDs");
+					return 1;
+				}
+			} else {
+				num = 0;
 			}
 
 			ids = uint16v_alloc(2 + num);
@@ -982,8 +1033,8 @@ int main(int argc, char **argv)
 				return 1;
 			}
 
-			ids->vec[0] = geteuid();
-			ids->vec[1] = real_gid;
+			ids->vec[0] = uid;
+			ids->vec[1] = gid;
 
 			for (i = 0; i < num; ++i)
 				ids->vec[i + 2] = gids[i];
